#include <crtdbg.h>
#include <stdio.h>
#include <ctype.h>
#include <vector>
#include <algorithm>
#include <utility>

void TrimSpace(std::string &r_s_string)
{
	int b = 0, e = r_s_string.length();
	while(e > 0 && isspace(r_s_string[e - 1]))
		-- e;
	while(b < e && isspace(r_s_string[b]))
		++ b;
	r_s_string.erase(e);
	r_s_string.erase(0, b);
}

bool ReadLine(FILE *p_fr, std::string &r_s_line, int &r_n_cur_line)
{
	while(!feof(p_fr)) {
		r_s_line.erase();
		try {
			for(int c = fgetc(p_fr); c != '\n' && c != EOF; c = fgetc(p_fr))
				r_s_line += c;
		} catch(std::bad_alloc&) {
			return false;
		}
		// read line

		++ r_n_cur_line;
		// line counter for file debugging

		if(r_s_line.find('#') != std::string::npos)
			r_s_line.erase(r_s_line.find('#'));
		// throw away line comment

		TrimSpace(r_s_line);
		// throw away begin / end whitespace

		if(!r_s_line.length())
			continue;
		// skip empty lines

		return true;
	}

	return false;
}

bool ReadFile(std::string &r_s_output, const char *p_s_filename)
{
	FILE *p_fr;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fr, p_s_filename, "rb"))
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fr = fopen(p_s_filename, "rb")))
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		return false;
	int n_length;
	if(fseek(p_fr, 0, SEEK_END) ||
	   (n_length = ftell(p_fr)) < 0 ||
	   fseek(p_fr, 0, SEEK_SET)) {
		fclose(p_fr);
		return false;
	}
	try {
		r_s_output.resize(n_length);
	} catch(std::bad_alloc&) {
		fclose(p_fr);
		return false;
	}
	if(fread(&r_s_output[0], sizeof(char), n_length, p_fr) != n_length) {
		fclose(p_fr);
		return false;
	}
	fclose(p_fr);
	return true;
}

void PrintHelp()
{
	printf("use: KindleFix [--max-chars-line 50] [--page-lines 28] [--content-lines 3]\n"
		   "               [--number-lines 1] -i input.txt [-o output.txt]\n\n"
		   "where:\n"
		   "--max-chars-line limits maximal number of characters per line (so that each line in file\n"
		   "\tfits a single line on your kindle dislay)\n"
		   "--page-lines sets number of lines per page (28 for kindle 3 and medium font size)\n"
		   "--content-lines is number of times to repeat the line with content (to avoid misalignment\n"
		   "\tafter seeking)\n"
		   "--number-lines is number of lines before and after content-lines that should contain\n"
		   "\tline number (just to ease of seeking in computer)\n"
		   "\n"
		   "number-lines * 2 + content-lines should be less than page-lines (will limit automatically)\n"
		   "\n"
		   "KindleFix is for people who dropped their Kingle and found themselves only with a small\n"
		   "portion of display still working. This modifies text files so that only one line is\n"
		   "displayed at a time, enabling simple reading even on those devices.\n"
		   "\n"
		   "The current version only repeats a single line, if someone found themselves in the need\n"
		   "of having more different lines displayed at a time (large portion of display still working),\n"
		   "just let me know.\n");
}

int main(int n_arg_num, const char **p_arg_list)
{
	int n_max_line_length = 50; // 
	int n_line_spacing = 28; // number of lines per screen
	int n_content_line_num = 3; // number of times the line with the content should repeat
	int n_line_number_lines_num = 1; // number of lines with line numbers surrounding the content

	const char *p_s_infile = 0;
	const char *p_s_outfile = 0;

	for(int i = 1; i < n_arg_num; ++ i) {
		if(!strcmp(p_arg_list[i], "--help") || !strcmp(p_arg_list[i], "-h")) {
			PrintHelp();
			return 0;
		} else if(i + 1 == n_arg_num) {
			fprintf(stderr, "error: argument \'%s\': unknown argument or"
				" needs a value (use --help to get help)\n", p_arg_list[i]);
			return -1;
		} else if(!strcmp(p_arg_list[i], "--max-chars-line")) {
			n_max_line_length = atoi(p_arg_list[++ i]);
			if(n_max_line_length <= 0) {
				fprintf(stderr, "warning: maximal line length set too low (defaulting to 50)\n");
				n_max_line_length = 50;
			}
		} else if(!strcmp(p_arg_list[i], "--page-lines")) {
			n_line_spacing = atoi(p_arg_list[++ i]);
			if(n_line_spacing <= 0) {
				fprintf(stderr, "warning: number of lines per page set too low (defaulting to 28)\n");
				n_line_spacing = 28;
			}
		} else if(!strcmp(p_arg_list[i], "--content-lines")) {
			n_content_line_num = atoi(p_arg_list[++ i]);
			if(n_content_line_num <= 0) {
				fprintf(stderr, "warning: number of content lines set too low (defaulting to 3)\n");
				n_content_line_num = 3;
			}
		} else if(!strcmp(p_arg_list[i], "--number-lines")) {
			n_line_number_lines_num = atoi(p_arg_list[++ i]);
			if(n_line_number_lines_num < 0) {
				fprintf(stderr, "warning: number of heading/trailing lines set too low (defaulting to 1)\n");
				n_line_number_lines_num = 1;
			}
		} else if(!strcmp(p_arg_list[i], "-i"))
			p_s_infile = p_arg_list[++ i];
		else if(!strcmp(p_arg_list[i], "-o"))
			p_s_outfile = p_arg_list[++ i];
		else {
			fprintf(stderr, "error: argument \'%s\': unknown argument"
				" (use --help to get help)\n", p_arg_list[i]);
			return -1;
		}
	}
	// parse commandline

	if(!p_s_infile) {
		PrintHelp();
		return 0;
	}
	// make sure input file is specified

	std::string s_outfile;
	if(!p_s_outfile) {
		try {
			s_outfile = p_s_infile;
			size_t n_pos;
			if((n_pos = s_outfile.rfind('.')) != std::string::npos)
				s_outfile.erase(n_pos);
			s_outfile += "_lines.txt";
			p_s_outfile = s_outfile.c_str();
		} catch(std::bad_alloc&) {
			fprintf(stderr, "error: not enough memory\n");
			return -1;
		}
	}
	// in case no outfile name given, make one from infile name

	std::string s_file;
	if(!ReadFile(s_file, p_s_infile)) {
		fprintf(stderr, "error: failed to read input file (\'%s\')\n", p_s_infile);
		return -1;
	}
	// read input file at once

	for(size_t n_pos = 0; (n_pos = s_file.find("\r\n\r\n", n_pos)) != std::string::npos;) {
		s_file[n_pos] = ' ';
		s_file[n_pos + 1] = '|';
		s_file[n_pos + 2] = ' ';
	}
	// replace spaces between paragraphs by pipes

	for(size_t i = 0, n = s_file.length(); i < n; ++ i) {
		int n_c = (unsigned char)s_file[i];
		if(strchr("\n\r\t", n_c))
			s_file[i] = ' ';
	}
	// replace special characters for spaces

	for(size_t n_pos = 0; (n_pos = s_file.find("  ", n_pos)) != std::string::npos;)
		s_file.erase(n_pos + 1, 1);
	// erase all the double spaces

	FILE *p_fw;
#if defined(_MSC_VER) && !defined(__MWERKS__) && _MSC_VER >= 1400
	if(fopen_s(&p_fw, p_s_outfile, "w")) {
#else //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
	if(!(p_fw = fopen(p_s_outfile, "w"))) {
#endif //_MSC_VER && !__MWERKS__ && _MSC_VER >= 1400
		fprintf(stderr, "error: failed to open output file for writing (\'%s\')\n", p_s_outfile);
		return -1;
	}
	// open output file for writing

	int n_line_no = 0;

	int n_content_line_before_num = n_content_line_num / 2;
	int n_content_line_after_num = n_content_line_num - 1 - n_content_line_before_num;
	_ASSERTE(n_content_line_before_num + 1 + n_content_line_after_num == n_content_line_num);
	_ASSERTE(n_content_line_num >= 1);
	_ASSERTE(n_content_line_before_num >= 0 && n_content_line_after_num >= 0);
	//int n_extra_lines = n_content_line_before_num + n_content_line_after_num;

	for(int i = 0; i < n_content_line_before_num; ++ i)
		fprintf(p_fw, "%d\n", ++ n_line_no);

	for(size_t b = 0, n = s_file.length(); b < n;) {
		int e = s_file.find(' ', b + 1);
		// find the next space

		_ASSERTE(e - b <= INT_MAX);
		if(int(e - b) >= n_max_line_length) {
			e = b + n_max_line_length;
			// have to split too long word
		} else {
			for(;;) {
				int e2 = s_file.find(' ', e + 1);
				if(e2 == std::string::npos) {
					e2 = n;
					_ASSERTE(e2 - b <= INT_MAX);
					if(int(e2 - b) >= n_max_line_length)
						break;
					e = e2;
					break;
				}
				// find the next space or the end of the string

				_ASSERTE(e2 - b <= INT_MAX);
				if(int(e2 - b) >= n_max_line_length)
					break;
				// too far? break.

				e = e2;
				// this is safe
			}
		}
		// split the sentence so as to form a line that is below n_max_line_length

		_ASSERTE(e - b <= INT_MAX);
		_ASSERTE(int(e - b) < n_max_line_length);

		for(int i = 0, m = n_line_spacing - (n_content_line_num - 1); i < m; ++ i) {
			if(i < n_line_number_lines_num || i >= m - n_line_number_lines_num)
				fprintf(p_fw, "%d\n", ++ n_line_no);
			else {
				++ n_line_no;
				fputc('\n', p_fw);
			}
		}

		for(int i = 0; i < n_content_line_num; ++ i) {
			++ n_line_no;
			for(int j = b; j < e; ++ j)
				fputc(s_file[j], p_fw);
			fputc('\n', p_fw);
		}
		b = e;
		// print that part of the file

		_ASSERTE(b == e); // it already is
		if(b < n && isspace(s_file[b]))
			++ b;
		_ASSERTE(!(b < n && isspace(s_file[b])));
		// skip space
	}

	if(ferror(p_fw)) {
		fclose(p_fw);
		fprintf(stderr, "error: error writing output file (disk full?)\n");
		return -1;
	}
	fclose(p_fw);

	return 0;
}
